% Function to perform reverse OOS extension using manifold charting
% Inputs: X (train data), Y (train embeddings), Ynew (test data), nn_size
%         (neighbourhood size for interpolation), nn (neighbourhood for PCA)
% Output: Xnew (test embeddings)

function Xnew = LQR(X,Y,Ynew,nn_size,nn,dim,lambda)

% Parameters
propagate_learning = 0;

Xnew = zeros(size(Ynew,1),size(X,2));

% Find reconstruction of each test point
for i = 1:size(Ynew,1)
    % Find the nearest neighbour of Ynew in Y
    target = Ynew(i,:);
    
    % Find the tangent plane at X(p,:)
    neighbourhood2 = knnsearch(Y,Ynew(i,:),'K',nn_size);
    neighbourhood = nn(i,:);
    [mappedX,mapping] = compute_mapping(X(neighbourhood,:),'PCA',dim);
    mX = (X(neighbourhood2,:)-repmat(mapping.mean,nn_size,1))*mapping.M;
    
    % Interpolate
    % Construct weights and compute OOS
    tol = 1e-3;
    W = zeros(nn_size,1);
    z = Y(neighbourhood2(1:nn_size),:)-repmat(target,nn_size,1); % shift ith pt to origin
    C = z*z';                                        % local covariance
    C = C + eye(nn_size,nn_size)*tol*trace(C);       % regularlization (K>D)
    W = C\ones(nn_size,1);                           % solve Cw=1
    W = W/sum(W);                  % enforce sum(w)=1
    
    % Tangent Space components
    mappedZ = W'*mX(1:nn_size,1:min(size(Y,2),size(mX,2)));
    mappedXnew = mappedZ;
    
    % Quadratic Regression
    for j = size(mappedZ,2)+1:size(mappedX,2)
        if (size(Y,2) == 1)
            A = [mappedX(:,1:size(Y,2)).^2 mappedX(:,1:size(Y,2)) ones(size(mappedX,1),1)];
            AA = A'*A;
            AA = AA + eye(2*size(Y,2)+1,2*size(Y,2)+1)*lambda;
            par = AA\A'*mappedX(:,j);
            newdim = [mappedZ.^2 mappedZ 1]*par;
        elseif (size(Y,2) == 2)
            A = [mappedX(:,1:size(Y,2)).^2 prod(mappedX(:,1:size(Y,2)),2) mappedX(:,1:size(Y,2)) ones(size(mappedX,1),1)];
            AA = A'*A;
            AA = AA + eye(2*size(Y,2)+2,2*size(Y,2)+2)*lambda;
            par = AA\A'*mappedX(:,j);
            newdim = [mappedZ.^2 prod(mappedZ,2) mappedZ 1]*par;
        else
            disp('Intrinsic Dimension of manifold must be less than 3.')
        end
        
        % Constrain
        if (mappedXnew(1) > max(mappedX(:,1)) || mappedXnew(1) < min(mappedX(:,1)))
            aa = max(mappedX(:,j));
            bb = min(mappedX(:,j));
            if(newdim > aa) newdim = aa;
            elseif (newdim < bb) newdim = bb;
            end
        end
        mappedXnew = [mappedXnew newdim];
        
    end
    Xnew(i,:) = (mappedXnew)*mapping.M' + mapping.mean;
end

% Modify the original data for future learning
if (propagate_learning)
    X = [X; Xnew(i,:)];
    Y = [Y; Ynew(i,:)];
end