% Script to generate a swiss-roll, and show Local Quadrature Reconstruction
% on it.

% Parameters
N = 1000;
rot = 5;
Ntest = 500;
Ntv = 20;
nn_size = 5;
nn_size2 = 18;
showid = 19;
lambda = 1;

% % Generate the Swiss Roll
tt = (3*pi/2)*(1+2*rand(1,N));  height = 5*rand(1,N);
data = [tt.*cos(tt); height; tt.*sin(tt)]';
Y = [tt;height]';

% Split into training and testing
testIdx = randsample(N,Ntest);
td = data(testIdx,:);
ty  = Y(testIdx,:);
trd = data(setdiff(1:size(data,1),testIdx),:);
ry = Y(setdiff(1:size(data,1),testIdx),:);
figure; 
h1 = scatter3(data(:,1),data(:,2),data(:,3),20,Y(:,1),'filled');
hold on; axis off;
view([10 5])
% print -dpsc ../../thesis_template/Figures/SwissRollX.eps
figure; scatter(Y(:,1),Y(:,2),20,Y(:,1),'filled');
axis off;
ylim([-2.5 7.5])
% print -dpsc ../../thesis_template/Figures/SwissRollY.eps

% % Compute the neighbourhoods
neighbourhood = knnsearch(ry,ty,'K',nn_size);
p = knnsearch(ry,ty);
neighbourhood2 = knnsearch(trd,trd(p,:),'K',nn_size2);

% % Compare overall performance
xtsnl = LQR(trd,ry,ty,nn_size,neighbourhood2,3,lambda);
xlin = linearReconstruction(trd,ry,ty,nn_size);
etsnl = sum((td-xtsnl).^2,2);
elin = sum((td-xlin).^2,2);
figure; plot(etsnl); hold on; plot(elin,'r')
% 
% % Show some tangent vectors
figure; scatter3(data(:,1),data(:,2),data(:,3),20,Y(:,1),'filled');
axis off; hold on;
for i = 1:Ntv
    [mappedX,mapping] = compute_mapping(trd(neighbourhood2(i,:),:),'PCA',3);
    a1 = [2 0 0]*mapping.M'+mapping.mean;
    a2 = [0 2 0]*mapping.M'+mapping.mean;
    a3 = [0 0 2]*mapping.M'+mapping.mean;
    h1 = plot3([mapping.mean(1) a1(1)],[mapping.mean(2) a1(2)],[mapping.mean(3) a1(3)],'k','LineWidth',1.5);
    h2 = plot3([mapping.mean(1) a2(1)],[mapping.mean(2) a2(2)],[mapping.mean(3) a2(3)],'k','LineWidth',1.5);
    h3 = plot3([mapping.mean(1) a3(1)],[mapping.mean(2) a3(2)],[mapping.mean(3) a3(3)],'LineWidth',1.5);
end
view([10 5])
hl = legend([h1 h3], 'Tangent Space Principal Components', 'Orthogonal Principal Component');
set(hl,'FontSize',15);
% print -dpsc ../../thesis_template/Figures/SwissRollTV.eps
% 
% % Show one example closely
figure; scatter3(trd(:,1),trd(:,2),trd(:,3),20,ry(:,1),'filled');
axis off; hold on;
view([10 5])
h1 = scatter3(td(showid,1),td(showid,2),td(showid,3),20,'k','filled');
h2 = scatter3(trd(neighbourhood2(showid,:),1),trd(neighbourhood2(showid,:),2),trd(neighbourhood2(showid,:),3),20,'k','LineWidth',1.5);
[mappedX,mapping] = compute_mapping(trd(neighbourhood2(showid,:),:),'PCA',3);
a1 = [1 0 0]*mapping.M'+mapping.mean;
a2 = [0 1 0]*mapping.M'+mapping.mean;
a3 = [0 0 1]*mapping.M'+mapping.mean;
plot3([mapping.mean(1) a1(1)],[mapping.mean(2) a1(2)],[mapping.mean(3) a1(3)],'k','LineWidth',1.5);
plot3([mapping.mean(1) a2(1)],[mapping.mean(2) a2(2)],[mapping.mean(3) a2(3)],'k','LineWidth',1.5);
plot3([mapping.mean(1) a3(1)],[mapping.mean(2) a3(2)],[mapping.mean(3) a3(3)],'LineWidth',1.5);
hl = legend([h1 h2], 'Test Point', 'Neighbourhood Points');
set(hl,'FontSize',15);
% print -dpsc ../../thesis_template/Figures/SwissRollXzoom.eps

% % Linear regression
mX = (trd(neighbourhood(showid,:),:)-repmat(mapping.mean,nn_size,1))*mapping.M;
tol = 1e-3;
W = zeros(nn_size,1);
z = ry(neighbourhood(showid,1:nn_size),:)-repmat(ty(showid,:),nn_size,1); % shift ith pt to origin
C = z*z';                                        % local covariance
C = C + eye(nn_size,nn_size)*tol*trace(C);       % regularlization (K>D)
W = C\ones(nn_size,1);                           % solve Cw=1
W = W/sum(W);                  % enforce sum(w)=1
mappedXnew = W'*mX;

% Quadratic Regression
lambda = 1;
A = [mappedX(:,1:2).^2 prod(mappedX(:,1:2),2) mappedX(:,1:2) ones(size(mappedX,1),1)];
AA = A'*A;
AA = AA + eye(6,6)*lambda;
par = AA\A'*mappedX(:,3);

newdim = [mappedXnew(1:2).^2 prod(mappedXnew(1:2),2) mappedXnew(1:2) 1]*par;
zlin = (xlin(showid,:)-mapping.mean)*mapping.M;

% Show the linear regression
figure;
h1 = plot(mX(:,1),mX(:,2),'.','MarkerSize',10);
for i = 1:nn_size
    text(mX(i,1)+0.1,mX(i,2),int2str(i),'FontSize',15);
end
hold on; grid on;
h2 = plot(mappedXnew(1),mappedXnew(2),'ro','LineWidth',2);
hl = legend([h1 h2],'Neighbourhood Projections','Interpolated Projection');
set(hl,'FontSize',15);
title('Tangent Space','FontSize',15)
% print -dpsc ../../thesis_template/Figures/SwissRollZnn.eps
figure;
h1 = plot(ry(neighbourhood(showid,:),1),ry(neighbourhood(showid,:),2),'.','MarkerSize',10);
for i = 1:nn_size
    text(ry(neighbourhood(showid,i),1)+0.01,ry(neighbourhood(showid,i),2),int2str(i),'FontSize',15);
end
hold on; grid on;
h2 = plot(ty(showid,1),ty(showid,2),'o','LineWidth',2);
hl = legend([h1 h2],'Neighbourhood Coordinates','Test Coordinates');
set(hl,'FontSize',15);
title('Latent Space','FontSize',15)
% print -dpsc ../../thesis_template/Figures/SwissRollYnn.eps

% Show the quadratic regression
figure;
h1 = plot3(mappedX(:,1),mappedX(:,2),mappedX(:,3),'.','MarkerSize',10);
hold on; grid on;
z1range = linspace(min([mappedX(:,1); mX(:,1)]),max([mappedX(:,1); mX(:,1)]),20);
z2range = linspace(min([mappedX(:,2); mX(:,2)]),max([mappedX(:,2); mX(:,2)]),20);
c = zeros(20,20);
for i = 1:20
    for j = 1:20
        c(j,i) = [z1range(i)^2 z2range(j)^2 z1range(i)*z2range(j) z1range(i) z2range(j) 1]*par;
    end
end
surf(z1range,z2range,c);
h3 = plot3(zlin(1),zlin(2),zlin(3),'rx','MarkerSize',10,'LineWidth',2);
h4 = plot3(mappedXnew(1),mappedXnew(2),newdim,'gx','MarkerSize',10,'LineWidth',2);
hl = legend([h1 h3 h4],'k_1-NN of nearest neighbour','Linear Regression','Quadratic Regression');
set(hl,'FontSize',15);
xlabel('Principal Component 1')
ylabel('Principal Component 2')
zlabel('Principal Component 3')
title('Quadratic Regression','FontSize',15)
% print -dpsc ../../thesis_template/Figures/SwissRollQuad.eps

% % Show overall reconstructions
figure; scatter3(td(:,1),td(:,2),td(:,3),20,ty(:,1),'LineWidth',1.5);
axis off;
view([0 0])
title('Original Data','FontSize',15)
% print -dpsc ../../thesis_template/Figures/SwissRollTD.eps
figure; scatter3(xtsnl(:,1),xtsnl(:,2),xtsnl(:,3),20,ty(:,1),'LineWidth',1.5);
axis off;
view([0 0])
xx = xlim;
yy = ylim;
zz = zlim;
title('Local Quadrature Reconstructions','FontSize',15)
% print -dpsc ../../thesis_template/Figures/SwissRollTS.eps
figure; scatter3(xlin(:,1),xlin(:,2),xlin(:,3),20,ty(:,1),'LineWidth',1.5);
axis off;
view([0 0])
xlim(xx);
ylim(yy);
zlim(zz);
title('Linear Reconstructions','FontSize',15)
% print -dpsc ../../thesis_template/Figures/SwissRollLin.eps
